@@ -10,7 +10,8 @@
 ** The IP masquerading functionality has been taken from openbsd.c and
 ** is distributed under the same copyright.
 **
-** Modifications Copyright (C) 1998-2006 Ryan McCabe <ryan@numb.org>
+** Modifications Copyright (c) 1998-2006 Ryan McCabe <ryan@numb.org>
+** Modifications Copyright (c) 2018-2019 Janik Rabe  <oidentd@janikrabe.com>
 */
 
 #define _WANT_UCRED
@@ -58,18 +59,18 @@
 
 #include <arpa/inet.h>
 
-#ifdef MASQ_SUPPORT
+#if MASQ_SUPPORT
 #	include <netinet/ip_compat.h>
 #	include <netinet/ip_fil.h>
 #	include <netinet/ip_nat.h>
 #endif
 
-#include <oidentd.h>
-#include <oidentd_util.h>
-#include <oidentd_inet_util.h>
-#include <oidentd_missing.h>
-#include <oidentd_options.h>
-#include <oidentd_masq.h>
+#include "oidentd.h"
+#include "util.h"
+#include "inet_util.h"
+#include "missing.h"
+#include "options.h"
+#include "masq.h"
 
 extern struct sockaddr_storage proxy;
 
@@ -79,7 +80,7 @@ extern struct sockaddr_storage proxy;
 
 #define N_TCB			0
 
-#ifdef MASQ_SUPPORT
+#if MASQ_SUPPORT
 #	define N_NATLIST	1
 #	define N_TOTAL		3
 #else
@@ -100,22 +101,23 @@ static struct kainfo {
 } *kinfo;
 
 /*
-** Open the kernel memory device
+** Open the kernel memory device.
+** Return 0 on success, or -1 with errno set.
 */
 
 int k_open(void) {
 	kinfo = xmalloc(sizeof(struct kainfo));
 
 	kinfo->kd = kvm_open(NULL, NULL, NULL, O_RDONLY, NULL);
-	if (kinfo->kd == NULL) {
+	if (!kinfo->kd) {
 		free(kinfo);
 		debug("kvm_open: %s", strerror(errno));
-		return (-1);
+		return -1;
 	}
 
 	kinfo->nl[N_TCB].n_name = "_tcb";
 
-#ifdef MASQ_SUPPORT
+#if MASQ_SUPPORT
 	if (opt_enabled(MASQ))
 		kinfo->nl[N_NATLIST].n_name = "_nat_instances";
 	else
@@ -128,15 +130,17 @@ int k_open(void) {
 		kvm_close(kinfo->kd);
 		free(kinfo);
 		debug("kvm_nlist: %s", strerror(errno));
-		return (-1);
+		return -1;
 	}
 
-#ifdef MASQ_SUPPORT
-	if (kinfo->nl[N_NATLIST].n_value == 0)
+#if MASQ_SUPPORT
+	if (opt_enabled(MASQ) && kinfo->nl[N_NATLIST].n_value == 0) {
+		o_log(LOG_CRIT, "NAT/IP masquerading support is unavailable");
 		disable_opt(MASQ);
+	}
 #endif
 
-	return (0);
+	return 0;
 }
 
 /*
@@ -148,10 +152,10 @@ static int getbuf(u_long addr, void *buf, size_t len) {
 
 	if (kvm_read(kinfo->kd, addr, buf, len) < 0) {
 		debug("getbuf: kvm_read: %s", strerror(errno));
-		return (-1);
+		return -1;
 	}
 
-	return (0);
+	return 0;
 }
 
 /*
@@ -170,8 +174,8 @@ static struct socket *getlist(	void *arg,
 	struct inpcb *pcbp = arg;
 	struct inpcb *head;
 
-	if (pcbp == NULL)
-		return (NULL);
+	if (!pcbp)
+		return NULL;
 
 	head = pcbp->inp_prev;
 
@@ -182,7 +186,7 @@ static struct socket *getlist(	void *arg,
 				pcbp->inp_fport == fport &&
 				pcbp->inp_lport == lport)
 			{
-				return (pcb.inp_socket);
+				return pcb.inp_socket;
 			}
 		}
 
@@ -191,12 +195,12 @@ static struct socket *getlist(	void *arg,
 			pcbp->inp_fport == fport &&
 			pcbp->inp_lport == lport)
 		{
-			return (pcbp->inp_socket);
+			return pcbp->inp_socket;
 		}
 	} while (pcbp->inp_next != head &&
 		getbuf((u_long) pcbp->inp_next, pcbp, sizeof(struct inpcb)) != -1);
 
-	return (NULL);
+	return NULL;
 }
 
 #else
@@ -213,7 +217,7 @@ static struct socket *getlist(	void *arg,
 	int alen;
 
 	if (remote->sa_family != local->sa_family)
-		return (NULL);
+		return NULL;
 	switch (remote->sa_family) {
 	case AF_INET:
 		faddr = (char *)&SIN4(remote)->sin_addr;
@@ -224,12 +228,12 @@ static struct socket *getlist(	void *arg,
 		laddr = (char *)&SIN6(local)->sin6_addr;
 		break;
 	default:
-		return (NULL);
+		return NULL;
 	}
 
 	head = pcbhead->lh_first;
-	if (head == NULL)
-		return (NULL);
+	if (!head)
+		return NULL;
 
 	do {
 		if (getbuf((u_long) head, &pcbp, sizeof(struct inpcb)) == -1)
@@ -241,7 +245,7 @@ static struct socket *getlist(	void *arg,
 				pcbp.inp_fport == fport &&
 				pcbp.inp_lport == lport)
 			{
-				return (pcbp.inp_socket);
+				return pcbp.inp_socket;
 			}
 		}
 
@@ -264,31 +268,32 @@ static struct socket *getlist(	void *arg,
 			pcbp.inp_fport == fport &&
 			pcbp.inp_lport == lport)
 		{
-			return (pcbp.inp_socket);
+			return pcbp.inp_socket;
 		}
 
 		head = pcbp.inp_list.le_next;
-	} while (head != NULL);
+	} while (head);
 
-	return (NULL);
+	return NULL;
 }
 
 #endif
 
 /*
-** System dependant initialization. Call only once!
-** On failure, return false.
+** System-dependent initialization; called only once.
+** Called before privileges are dropped.
+** Returns non-zero on failure.
 */
 
-bool core_init(void) {
-	return (true);
+int core_init(void) {
+	return 0;
 }
 
 /*
 ** Return the UID of the connection owner
 */
 
-static int get_user(	in_port_t lport,
+static uid_t get_user(	in_port_t lport,
 				in_port_t fport,
 				struct sockaddr_storage *laddr,
 				struct sockaddr_storage *faddr)
@@ -303,13 +308,13 @@ static int get_user(	in_port_t lport,
 	struct inpcbhead tcb;
 #endif
 	kp = kvm_getprocs(kinfo->kd, KERN_PROC_ALL, 0, &nentries);
-	if (kp == NULL) {
+	if (!kp) {
 		debug("kvm_getprocs: %s", strerror(errno));
-		return (-1);
+		return MISSING_UID;
 	}
 
 	if (getbuf(kinfo->nl[N_TCB].n_value, &tcb, sizeof(tcb)) == -1)
-		return (-1);
+		return MISSING_UID;
 
 #ifdef _HAVE_OLD_INPCB
 	tcb.inp_prev = (struct inpcb *) kinfo->nl[N_TCB].n_value;
@@ -319,22 +324,22 @@ static int get_user(	in_port_t lport,
 				(struct sockaddr *)laddr,
 				(struct sockaddr *)faddr);
 
-	if (sockp == NULL)
-		return (-1);
+	if (!sockp)
+		return MISSING_UID;
 
 	/*
 	** Locate the file descriptor that has the socket in question
 	** open so that we can get the 'ucred' information
 	*/
-	for (i = 0 ; i < nentries ; i++) {
+	for (i = 0; i < nentries; ++i) {
 
-		if ( kp[i].ki_fd != NULL) {
+		if (kp[i].ki_fd) {
 			int j;
 			int ret;
 			struct filedesc pfd;
 			struct file **ofiles;
 			if (getbuf((u_long) kp[i].ki_fd, &pfd, sizeof(pfd)) == -1)
-				return (-1);
+				return MISSING_UID;
 			ofiles = xmalloc(pfd.fd_nfiles * sizeof(struct file *));
 
 			ret = getbuf((u_long) pfd.fd_ofiles, ofiles,
@@ -342,19 +347,19 @@ static int get_user(	in_port_t lport,
 
 			if (ret == -1) {
 				free(ofiles);
-				return (-1);
+				return MISSING_UID;
 			}
 
-			for (j = 0 ; j < pfd.fd_nfiles ; j++) {
+			for (j = 0; j < pfd.fd_nfiles; ++j) {
 				struct file ofile;
 
-				if (ofiles[j] == NULL)
+				if (!ofiles[j])
 					continue;
 
 				ret = getbuf((u_long) ofiles[j], &ofile, sizeof(struct file));
 				if (ret == -1) {
 					free(ofiles);
-					return (-1);
+					return MISSING_UID;
 				}
 
 				if (ofile.f_count == 0)
@@ -366,30 +371,36 @@ static int get_user(	in_port_t lport,
 
 					free(ofiles);
 
-					return (kp[i].ki_ruid);
+					return kp[i].ki_ruid;
 				}
 			}
 
 			free(ofiles);
 		}
-		
+
 	}
 
-	return (-1);
+	return MISSING_UID;
 }
 
-int get_user4(	in_port_t lport,
+/*
+** Returns the UID of the owner of an IPv4 connection,
+** or MISSING_UID on failure.
+*/
+
+uid_t get_user4(	in_port_t lport,
 				in_port_t fport,
 				struct sockaddr_storage *laddr,
 				struct sockaddr_storage *faddr)
 {
-	return (get_user(lport, fport, laddr, faddr));
+	return get_user(lport, fport, laddr, faddr);
 }
 
-#ifdef MASQ_SUPPORT
+#if MASQ_SUPPORT
 
 /*
-** Check ident requests for NAT connection
+** Handle a request to a host that's IP masquerading through us.
+** Returns non-zero on failure.
 */
 
 int masq(	int sock,
@@ -409,14 +420,15 @@ int masq(	int sock,
 	*/
 
 	if (faddr->ss_family != AF_INET || laddr->ss_family != AF_INET)
-		return (-1);
+		return -1;
 
 	if (getbuf(kinfo->nl[N_NATLIST].n_value, &np, sizeof(np)) == -1)
-		return (-1);
+		return -1;
 
-	for (; np != NULL ; np = nat.nat_next) {
-		int ret;
+	for (; np; np = nat.nat_next) {
 		in_port_t masq_lport;
+		in_port_t masq_fport;
+		int retm;
 
 		if (getbuf((u_long) np, &nat, sizeof(nat)) == -1) {
 			debug("getbuf: %s", strerror(errno));
@@ -449,15 +461,21 @@ int masq(	int sock,
 		lport = ntohs(lport);
 		fport = ntohs(fport);
 		masq_lport = ntohs(nat.nat_inport);
+		masq_fport = ntohs(nat.nat_outport);
 
 		sin_setv4(nat.nat_inip.s_addr, &ss);
 
-		if (opt_enabled(FORWARD)) {
-			ret = fwd_request(sock, lport, masq_lport, fport, &ss);
+		retm = find_masq_entry(&ss, user, sizeof(user), os, sizeof(os));
 
-			if (ret == 0)
-				return (0);
-			else {
+		if (opt_enabled(FORWARD) && (retm != 0 || !opt_enabled(MASQ_OVERRIDE))) {
+			int retf;
+
+			retf = fwd_request(sock, lport, masq_lport, fport, masq_fport, &ss);
+
+			if (retf == 0) {
+				if (retm != 0)
+					return 0;
+			} else {
 				char ipbuf[MAX_IPLEN];
 
 				get_ip(&ss, ipbuf, sizeof(ipbuf));
@@ -467,36 +485,40 @@ int masq(	int sock,
 			}
 		}
 
-		ret = find_masq_entry(&ss, user, sizeof(user), os, sizeof(os));
-		if (ret == 0) {
+		if (retm == 0) {
 			char ipbuf[MAX_IPLEN];
 
-			sockprintf(sock, "%d , %d : USERID : %s : %s\r\n",
+			sockprintf(sock, "%d,%d:USERID:%s:%s\r\n",
 				lport, fport, os, user);
 
 			get_ip(faddr, ipbuf, sizeof(ipbuf));
 
-			o_log(NORMAL,
+			o_log(LOG_INFO,
 				"[%s] (NAT) Successful lookup: %d , %d : %s",
 				ipbuf, lport, fport, user);
 
-			return (0);
+			return 0;
 		}
 	}
 
-	return (-1);
+	return -1;
 }
 
 #endif
 
-#ifdef WANT_IPV6
+#if WANT_IPV6
 
-int get_user6(	in_port_t lport,
+/*
+** Returns the UID of the owner of an IPv6 connection,
+** or MISSING_UID on failure.
+*/
+
+uid_t get_user6(	in_port_t lport,
 				in_port_t fport,
 				struct sockaddr_storage *laddr,
 				struct sockaddr_storage *faddr)
 {
-	return (get_user(lport, fport, laddr, faddr));
+	return get_user(lport, fport, laddr, faddr);
 }
 
 #endif
