@@ -11,7 +11,7 @@
 ** This code was taken from OpenSSH. It is distributed
 ** under a two-term BSD license.
 **
-** Modifications are Copyright (C) 2001-2006 Ryan McCabe <ryan@numb.org>
+** Modifications Copyright (c) 2001-2006 Ryan McCabe <ryan@numb.org>
 */
 
 #ifndef _GNU_SOURCE
@@ -37,7 +37,7 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
-#include <oidentd_missing.h>
+#include "missing.h"
 
 #ifndef HAVE_GAI_STRERROR
 
